####################################################################################################
#  Bayesian Estimation of Sparsely Heterogenous Panel Models
#  Empirical Analysis, unbalanced Panel
#####################################################################################################

#* Version: unbalanced panels, M2

####################################################################################################
#                             Part 0: INTRO
####################################################################################################
cd()

wd = "$(pwd())/Dropbox/SparseHetero/Software/ReplicationFile_v231019/Empirics/Empirics_Unbalanced/"

version = "_unbalanced"

wd_data = string(wd, "data", version, "/")

# create folder: Results (csv etc.)
wd_results = string(wd, "Results/")
if !isdir(wd_results)
   mkdir(wd_results)
end

# create folder: estimation output
wd_estimate = string(wd, "Estimates/")
if !isdir(wd_estimate)
   mkdir(wd_estimate)
end

# create folder: estimation output
wd_output = string(wd_estimate, "M2/")
if !isdir(wd_output)
   mkdir(wd_output)
end

# create temp folder for jld2 outout for hist
wd_jld = string(wd_output, "jld_temp/")
if !isdir(wd_jld)
   mkdir(wd_jld)
end


# create folder: Plots
wd_plots = string(wd, "Plots/")
if !isdir(wd_plots)
   mkdir(wd_plots)
end

# create hist folder
wd_hist = string(wd_plots, "Hist/")
if !isdir(wd_hist)
   mkdir(wd_hist)
end

# create folder: Plots
wd_hist = string(wd_hist, "M2/")
if !isdir(wd_hist)
   mkdir(wd_hist)
end

cd(wd)

# load packages (install them with Pkg.add() if needed)
using JLD2, FileIO, MutableNamedTuples, Random, Distributions, LinearAlgebra, SharedArrays, SpecialFunctions, DelimitedFiles, MAT
using SkipNan # deal wirh NANs in dataset
using ProgressBars # for fun

# load subfunction files
include("functions_emp_unbalanced.jl")
include("functions_helpers.jl")

####################################################################################################
#                             Part 1: INITALIZATION
####################################################################################################

## Data settings
# starting year
yr0 = 67
# end year (= 92(96) if last year = 1993(1997))
yr1 = 96

## MCMC settings
# number of MCMC draws
nMCMC = 10000
burnin = Int(floor(0.5*nMCMC))+2

# target acceptance rate of RWMH
opt_ar = 0.44

####################################################################################################
#                             Part 2: ESTIMATION
####################################################################################################
# helper functions
IG_nu(mean, sd) = 4+2*mean^2/sd^2
IG_tau(mean, sd) = 2*mean*(1+mean^2/sd^2)

##########################
# GLP, spike-and-slab
#########################

# Hyperparameter in the prior of v_δ^α (Inverse Wishart)
nu_vdelta_alpha  = 5.05
# Psi_vdelta_alpha = Matrix(1.0I, 2, 2)
Psi_vdelta_alpha = [0.5 0; 0 0.1]

# Hyperparameter in the prior of v_δ^ρ (Inverse Gamma)
prior_mean_delta_rho = 0.5^2 # 0.01^2
prior_sd_delta_rho   = 0.1 # 0.5
nu_vdelta_rho  = IG_nu(prior_mean_delta_rho, prior_sd_delta_rho)
tau_vdelta_rho = IG_tau(prior_mean_delta_rho, prior_sd_delta_rho)

# Hyperparameter in the prior of the implied variance v_δ^σ_u (Inverse Gamma)
prior_mean_vdelta_sigma_u = 1
prior_sd_vdelta_sigma_u   = 0.5
nu_vdelta_sigma_u  = IG_nu(prior_mean_vdelta_sigma_u, prior_sd_vdelta_sigma_u)
tau_vdelta_sigma_u = IG_tau(prior_mean_vdelta_sigma_u, prior_sd_vdelta_sigma_u)

# Hyperparameter in the prior of the implied variance v_δ^σ_ϵ (Inverse Gamma)
prior_mean_vdelta_sigma_e = 1
prior_sd_vdelta_sigma_e   = 0.5
nu_vdelta_sigma_e  = IG_nu(prior_mean_vdelta_sigma_e, prior_sd_vdelta_sigma_e)
tau_vdelta_sigma_e = IG_tau(prior_mean_vdelta_sigma_e, prior_sd_vdelta_sigma_e)

# Hyperparameter in the prior of σ_u (Inverse Gamma)
prior_mean_sigma_u = 0.05
prior_sd_sigma_u   = 0.05
nu_sigma_u  = IG_nu(prior_mean_sigma_u, prior_sd_sigma_u)
tau_sigma_u = IG_tau(prior_mean_sigma_u, prior_sd_sigma_u)

prior_mean_sigma_e = 0.05
prior_sd_sigma_e   = 0.05
nu_sigma_e  = IG_nu(prior_mean_sigma_e, prior_sd_sigma_e)
tau_sigma_e = IG_tau(prior_mean_sigma_e, prior_sd_sigma_e)

# Hyperparameter in the prior of the variance s0 (Inverse Gamma)
prior_mean_sigma_s0 = 0.05
prior_sd_sigma_s0   = 0.05
nu_sigma_s0  = IG_nu(prior_mean_sigma_s0, prior_sd_sigma_s0)
tau_sigma_s0 = IG_tau(prior_mean_sigma_s0, prior_sd_sigma_s0)

# # hyperparameters
lambda = (mu_alpha = [0,0], valpha = [1 0; 0 1], mu_rho = 0.8, vrho = 1, 
         nu_vdelta_alpha = nu_vdelta_alpha, Psi_vdelta_alpha = Psi_vdelta_alpha,
         nu_vdelta_rho = nu_vdelta_rho, tau_vdelta_rho = tau_vdelta_rho,
         nu_vdelta_sigma_u = nu_vdelta_sigma_u, tau_vdelta_sigma_u = tau_vdelta_sigma_u,
         nu_vdelta_sigma_e = nu_vdelta_sigma_e, tau_vdelta_sigma_e = tau_vdelta_sigma_e,
         nu_sigma_u = nu_sigma_u, tau_sigma_u = tau_sigma_u, 
         nu_sigma_e = nu_sigma_e, tau_sigma_e = tau_sigma_e, 
         a = 1, b = 1,
         mu_s0 = 0, v_s0 = 0.05,
         nu_sigma_s0 = nu_sigma_s0, tau_sigma_s0 = tau_sigma_s0)

seed = 0606

T0 = 67
T = 96 - T0 + 1

# load the data
raw = readdlm(string(wd_data, "unbalanced_data_incl_NA.txt"), ' ')
# replace NA with NaN
raw = replace(raw, "NA" => NaN) 

id = unique(raw[:,1])
N  = length(id)

Y = zeros(T, N)
h = zeros(T, N)

for i = 1:N
   selected = (abs.(raw[:,1] .- id[i]) .< 0.05)
   Y[:,i] = raw[selected, 3]
   h[:,i] = raw[selected, 5]
end

# remove the last observation
loc_last_obs = mapslices(find_last_nonNA, Y, dims = 1)[1,:]

y_all = copy(Y)
h_all = copy(h)
for i in 1:N
   y_all[loc_last_obs[i],i] = NaN
   h_all[loc_last_obs[i],i] = NaN
end

y_all = y_all[1:end-1,:]
h_all = h_all[1:end-1,:]

# variables in age-earning portfolio (in-sample)
H = zeros(T-1, N, 2)
H[:,:,1] .= 1
H[:,:,2] = h_all/10

loc_nan = (y_all .=== NaN)
H[loc_nan,1] .= NaN


##########################
# spike-and-slab
#########################

seed = 100*T

post_mean_ss_hetsk, post_draw_ss_hetsk = est_M2_unb(y_all, H, lambda, opt_ar, nMCMC, seed);


# OOS forecast
h_last = [h[loc_last_obs[i],i] for i in 1:N]
H_last = [ones(N,1) h_last/10];

alpha = post_mean_ss_hetsk.alpha_hat
delta_alpha = post_mean_ss_hetsk.delta_alpha_hat
delta_rho = post_mean_ss_hetsk.delta_rho_hat
rho = post_mean_ss_hetsk.rho_hat
s_last = [post_mean_ss_hetsk.s_hat[loc_last_obs[i],i] for i in 1:N]

fcst = sum(H_last .* (alpha' .+ delta_alpha), dims = 2) + s_last .* (rho .+ delta_rho)

# save lambda_i and rho_i in the case folder
file = matopen(string(wd_output, "results.mat"), "w")
post_mean_ss_hetsk_mat =(; post_mean_ss_hetsk..., fcst = fcst) # add fcst into final output
write(file, "post_mean", post_mean_ss_hetsk_mat)
close(file)

# save everything for plotting histograms
jldsave(wd_jld * "output.jld2"; post_mean_ss_hetsk = post_mean_ss_hetsk, post_draw_ss_hetsk = post_draw_ss_hetsk)